package net.yale.cs.p4p.p2p.azureus.client.data.dht;

import java.util.LinkedList;

import net.yale.cs.p4p.p2p.azureus.client.P4PPlugin;

import org.gudy.azureus2.plugins.ddb.DistributedDatabase;
import org.gudy.azureus2.plugins.ddb.DistributedDatabaseEvent;
import org.gudy.azureus2.plugins.ddb.DistributedDatabaseKey;
import org.gudy.azureus2.plugins.ddb.DistributedDatabaseListener;
import org.gudy.azureus2.plugins.ddb.DistributedDatabaseValue;

public class DHTTest extends Thread {

	private static final boolean IS_SERVER = true;

	private static final int START_DELTA = 60 * 1000;
	private static final int INTERVAL = 1 * 1000;
	private static final String BASE_KEY = DHTTest.class.getName();
	private static final long START_TIME = System.currentTimeMillis();

	private class DHTListener implements DistributedDatabaseListener {
		private long startTime_ = System.currentTimeMillis();
		private Operation op_;
		private int valuesRead_ = 0;
		private int valuesWritten_ = 0;
		public DHTListener(Operation op) {
			op_ = op;
		}
		private void log(String msg) {
			plugin_.getLogger().log(op_.getName() + " (" + ((System.currentTimeMillis() - startTime_) / 1000) + " secs) : " + msg);
		}
		public void event(DistributedDatabaseEvent event) {
			try {
				switch (event.getType()) {
				case DistributedDatabaseEvent.ET_OPERATION_STARTS:
					log("starting");
					break;
				case DistributedDatabaseEvent.ET_OPERATION_COMPLETE:
					log("finished: " + valuesRead_ + " read; " + valuesWritten_ + " written; " + (event.getKeyStats() != null ? event.getKeyStats().getEntryCount() : -1) + " entry count");
					break;
				case DistributedDatabaseEvent.ET_OPERATION_TIMEOUT:
					log("timeout");
					break;
				case DistributedDatabaseEvent.ET_VALUE_WRITTEN:
					log("wrote value: value = " + event.getValue().getValue(String.class) + "; version = " + event.getValue().getVersion() + "; contact = " + event.getValue().getContact().getAddress().toString());
					++valuesWritten_;
					break;
				case DistributedDatabaseEvent.ET_VALUE_READ:
					log("read value: value = " + event.getValue().getValue(String.class) + "; version = " + event.getValue().getVersion() + "; contact = " + event.getValue().getContact().getAddress().toString());
					++valuesRead_;
					break;
				case DistributedDatabaseEvent.ET_VALUE_DELETED:
					log("value deleted");
					break;
				case DistributedDatabaseEvent.ET_KEY_STATS_READ:
					log("key stats read: " + event.getKeyStats().getEntryCount() + " total entries");
					break;
				}				
			} catch (Exception e) {
				log("failed to log event: " + e);
			}
		}
	}

	private abstract class Operation {
		private String name_;
		private long deadline_;
		public String getName() { return name_ + "(" + getOpName() + ")"; }
		public String getKey() { return BASE_KEY + name_; }
		public boolean isExpired() { return START_TIME + deadline_ <= System.currentTimeMillis(); }
		public Operation(String name, long deadline) {
			name_ = name;
			deadline_ = deadline;
		}
		public abstract String getOpName();
		public abstract void start() throws Exception;
	}

	public class PutOperation extends Operation {
		private int count_;
		public int getCount() { return count_; }
		public PutOperation(String name, long deadline, int count) {
			super(name, deadline);
			count_ = count;
		}
		public String getOpName() { return "put"; }
		public void start() throws Exception {
			DistributedDatabaseKey key = db_.createKey(getKey());
			DistributedDatabaseValue[] val = new DistributedDatabaseValue[count_];
			for (int i = 0; i < val.length; ++i)
				val[i] = db_.createValue(Integer.toString(i));
			db_.write(new DHTListener(this), key, val);
		}
	}

	public class GetOperation extends Operation {
		public GetOperation(String name, long deadline) {
			super(name, deadline);
		}
		public String getOpName() { return "get"; }
		public void start() throws Exception {
			DistributedDatabaseKey key = db_.createKey(getKey());
			db_.read(new DHTListener(this), key, 600 * 1000);
		}
	}
	
	public class GetStatsOperation extends Operation {
		public GetStatsOperation(String name, long deadline) {
			super(name, deadline);
		}
		public String getOpName() { return "getstats"; }
		public void start() throws Exception {
			DistributedDatabaseKey key = db_.createKey(getKey());
			db_.readKeyStats(new DHTListener(this), key, 600 * 1000);
		}
	}

	private P4PPlugin plugin_;
	private DistributedDatabase db_;

	private LinkedList<Operation> operationList_ = new LinkedList<Operation>();

	public DHTTest(P4PPlugin plugin) {
		plugin_ = plugin;
		db_ = plugin_.getPluginInterface().getDistributedDatabase();

		if (IS_SERVER) {
			operationList_.add(new PutOperation("put002-get", START_DELTA + 1*1000, 2));
			operationList_.add(new PutOperation("put500-get", START_DELTA + 120*1000, 500));
		} else {
			operationList_.add(new PutOperation("put002-get", START_DELTA + 1*1000, 3));
			operationList_.add(new GetOperation("put002-get", START_DELTA + 240*1000));
			operationList_.add(new GetOperation("put500-get", START_DELTA + 360*1000));
		}
	}

	private void process() throws Exception {

		if (operationList_.isEmpty())
			return;

		if (!operationList_.peekFirst().isExpired())
			return;
		
		plugin_.getLogger().log("starting operation " + operationList_.peekFirst().getName());
		operationList_.removeFirst().start();
	}

	public void run() {
		while (!interrupted()) {
			try {
				process();
				sleep(INTERVAL);
			} catch (InterruptedException e) {
				break;
			} catch (Exception e) {
				plugin_.getLogger().log("DHT operation failed", e);
			}
		}
	}
	
}
