package net.yale.cs.p4p.p2p.azureus.client.data.selection;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Random;

import net.yale.cs.p4p.p2p.azureus.client.P4PPlugin;
import net.yale.cs.p4p.p2p.azureus.client.data.PDistCache;
import net.yale.cs.p4p.p2p.azureus.client.data.PeerCache;
import net.yale.cs.p4p.p2p.azureus.client.data.TorrentState;
import net.yale.cs.p4p.p2p.azureus.client.data.Torrents;
import net.yale.cs.p4p.p2p.azureus.client.data.TorrentState.PeerSnapshot;
import net.yale.cs.p4p.p2p.azureus.client.hacks.DiscoveredPeersHack;
import net.yale.cs.p4p.p2p.azureus.client.hacks.PeerClassifierHack;

import org.gudy.azureus2.plugins.PluginConfig;
import org.gudy.azureus2.plugins.download.Download;
import org.gudy.azureus2.plugins.peers.Peer;
import org.gudy.azureus2.plugins.peers.PeerDescriptor;
import org.gudy.azureus2.plugins.peers.PeerManager;
import org.gudy.azureus2.plugins.torrent.Torrent;

public abstract class PeerSelectionBase {

	public static final ScheduledPeerInfo EMPTY_SCHEDULED_PEER_INFO = new ScheduledPeerInfo();

	public static final class ScheduledPeerInfo {
		private static final Collection<String> EMPTY_LOG_TAGS = Collections.emptySet();

		private Collection<String> logTags_;
		public Collection<String> getLogTags() { return logTags_; }

		public ScheduledPeerInfo(Collection<String> logTags) {
			logTags_ = Collections.unmodifiableCollection(new ArrayList<String>(logTags));
		}
		
		public ScheduledPeerInfo() {
			logTags_ = EMPTY_LOG_TAGS;
		}
		
		public ScheduledPeerInfo(ScheduledPeerInfo info, Collection<String> logTags) {
			logTags_ = new ArrayList<String>(info.getLogTags().size() + logTags.size());
			logTags_.addAll(info.getLogTags());
			logTags_.addAll(logTags);
			logTags_ = Collections.unmodifiableCollection(logTags_);
		}
	}
	
	public static final class ScheduledPeer {
		
		private PeerDescriptor peer_;
		public PeerDescriptor getPeer() { return peer_; }

		private ScheduledPeerInfo schedInfo_;
		public ScheduledPeerInfo getScheduleInfo() { return schedInfo_; }

		public ScheduledPeer(PeerDescriptor peer, ScheduledPeerInfo schedInfo) {
			peer_ = peer;
			schedInfo_ = schedInfo;
		}
		
		public ScheduledPeer(PeerDescriptor peer) {
			this(peer, EMPTY_SCHEDULED_PEER_INFO);
		}
	}

	public static final class ScheduleResult {
		private static final Collection<Peer> EMPTY_DISCONNECT_PEERS = Collections.emptySet();
		
		private List<ScheduledPeer> scheduledPeers_;
		public List<ScheduledPeer> getScheduledPeers() { return scheduledPeers_; }

		private Collection<Peer> disconnectPeers_;
		public Collection<Peer> getDisconnectPeers() { return disconnectPeers_; }
		
		public ScheduleResult(List<ScheduledPeer> scheduledPeers, Collection<Peer> disconnectPeers) {
			scheduledPeers_ = scheduledPeers;
			disconnectPeers_ = disconnectPeers;
		}

		public ScheduleResult(List<ScheduledPeer> scheduledPeers) {
			this(scheduledPeers, EMPTY_DISCONNECT_PEERS);
		}
	}
	
	private P4PPlugin plugin_;
	public final P4PPlugin getPlugin() { return plugin_; }

	private PeerCache peerCache_;
	public final PeerCache getPeerCache() { return peerCache_; }

	public PDistCache pdistCache_;
	public final PDistCache getPDistCache() { return pdistCache_; }

	private Torrents torrents_;
	public final Torrents getTorrents() { return torrents_; }

	private Random rng_ = new Random();
	public final Random getRNG() { return rng_; } 

	public PeerSelectionBase(P4PPlugin plugin) {
		plugin_ = plugin;
		peerCache_ = plugin_.getPeerCache();
		pdistCache_ = plugin_.getPDistCache();
		torrents_ = plugin_.getTorrents();
	}

	public final int getMaxSelectedPeers() {
		int res = plugin_.getPluginInterface().getPluginconfig().getCoreIntParameter(PluginConfig.CORE_PARAM_INT_MAX_CONNECTIONS_PER_TORRENT);
		if (res > 0)
			return res;
		
		res = plugin_.getPluginInterface().getPluginconfig().getCoreIntParameter(PluginConfig.CORE_PARAM_INT_MAX_CONNECTIONS_GLOBAL);
		if (res > 0)
			return res;

		return plugin_.getConfig().getMaxSelectedPeers();
	}

	public final Peer selectOptimisticUnchoke(Download download) {
		Torrent torrent = download.getTorrent();
		if (torrent == null)
			return null;

		PeerManager peerManager = download.getPeerManager();
		if (peerManager == null)
			return null;

		Peer[] peers = peerManager.getPeers();
		if (peers == null)
			return null;

		/* Get the set of peers we are currently choking */
		Collection<Peer> chokedPeers = new ArrayList<Peer>();
		for (Peer peer : peers)
			if (peer.isChoking() && peer.getState() == Peer.TRANSFERING && peer.isInteresting() && !peer.isSeed())
				chokedPeers.add(peer);
		
		if (chokedPeers.isEmpty())
			return null;

		return selectOptimisticUnchoke(download, torrent, chokedPeers);
	}

	protected abstract Peer selectOptimisticUnchoke(Download download, Torrent torrent, Collection<Peer> peers);

	public final void updatePeers(Download download) {
		Torrent torrent = download.getTorrent();
		if (torrent == null)
			return;

		PeerManager peerManager = download.getPeerManager();
		if (peerManager == null)
			return;

		TorrentState torrentState = getTorrents().getTorrent(torrent);

		/* Get the peers available to be scheduled */
		PeerSnapshot schedPeers = torrentState.getSchedulablePeers(download.getStats(), DiscoveredPeersHack.getDiscoveredPeersCount(peerManager));
		if (schedPeers.isEmpty()) {
			getPlugin().getLogger().log(torrent, "No available peers");
			return;
		}

		getPlugin().getLogger().log(torrent, schedPeers.size() + " total peers");

		/* Get the new list of peers */
		int count = getMaxSelectedPeers();
		getPlugin().getLogger().log(torrent, "Selecting up to " + count + " peers");
		ScheduleResult schedResult = selectPeers(download, torrent, peerManager, schedPeers, count);

		/* Set these peers as queued these peers */
		List<ScheduledPeer> queuedPeers = torrentState.peersQueued(schedResult.getScheduledPeers());
		
		/* Extract peer descriptors */
		List<PeerDescriptor> queuedPeerDescs = new ArrayList<PeerDescriptor>(queuedPeers.size());
		for (ScheduledPeer queuedPeer : queuedPeers) {
			getPlugin().getLogger().log(torrent, "Scheduling peer " + TorrentState.peerDescriptorToString(queuedPeer.getPeer()));
			queuedPeerDescs.add(queuedPeer.getPeer());
		}

		/* Apply the new list of peers into Vuze and get the replaced ones */
		getPlugin().getLogger().log(torrent, "Adding " + queuedPeers.size() + " peers to connection queue");
		List<PeerDescriptor> replacedPeers = DiscoveredPeersHack.setDiscoveredPeers(peerManager, queuedPeerDescs);
		getPlugin().getLogger().log(torrent, "Counted " + DiscoveredPeersHack.getDiscoveredPeersCount(peerManager) + " peers in connection queue");

		/* In case the replaced peers were new peers (they shouldn't be), add them but be sure this doesn't count as an update */
		if (!replacedPeers.isEmpty()) {
			if (getPlugin().getLogger().isEnabled()) {
				getPlugin().getLogger().log(torrent, "Replaced " + replacedPeers.size() + " peers in queue");
				for (PeerDescriptor peer : replacedPeers)
					getPlugin().getLogger().log(torrent, "Replaced " + TorrentState.peerDescriptorToString(peer));
			}
			
			peerCache_.addPeers(replacedPeers);
			torrentState.peersDiscovered(replacedPeers, false);
		}
		
		/* Disconnect from the desired peers */
		for (Peer disconnectPeer : schedResult.getDisconnectPeers()) {
			if (PeerClassifierHack.isAzureusIP(disconnectPeer.getIp()))
				continue;
			peerManager.removePeer(disconnectPeer);
		}
	}

	protected abstract ScheduleResult selectPeers(Download download, Torrent torrent, PeerManager peerManager, PeerSnapshot schedPeers, int count);

	protected final <T> List<T> getRandomElements(ArrayList<T> array, int count) {
		ArrayList<T> result = new ArrayList<T>(count);
		for (int i = 0; i < count && !array.isEmpty(); ++i)
			result.add(array.remove(rng_.nextInt(array.size())));
		return result;
	}

	protected final <T> T getRandomElement(ArrayList<T> array) {
		if (array.isEmpty())
			return null;
		return array.remove(rng_.nextInt(array.size()));
	}

	protected final void fillScheduledPeers(List<PeerDescriptor> peers, List<ScheduledPeer> schedPeers, ScheduledPeerInfo baseSchedInfo) {
		for (PeerDescriptor peer : peers)
			schedPeers.add(new ScheduledPeer(peer, new ScheduledPeerInfo(baseSchedInfo, Collections.singleton("qpos=" + schedPeers.size()))));
	}

	protected void fillRandomPeers(PeerSnapshot schedPeers, int maxScheduled, ScheduledPeerInfo baseSchedInfo, List<ScheduledPeer> result) {

		/* Fill with 'available' peers */
		fillScheduledPeers(
			getRandomElements(new ArrayList<PeerDescriptor>(schedPeers.getAvailPeers()), maxScheduled),
			result,
			new ScheduledPeerInfo(baseSchedInfo, Collections.singleton("available")));
	}
	
	protected ScheduleResult selectPeersRandom(PeerSnapshot schedPeers, int count, ScheduledPeerInfo baseSchedInfo) {
		ArrayList<ScheduledPeer> selectedPeers = new ArrayList<ScheduledPeer>(count);
		fillRandomPeers(schedPeers, count, baseSchedInfo, selectedPeers);
		return new ScheduleResult(selectedPeers);
	}

}
