package net.yale.cs.p4p.p2p.azureus.client.experiment;

import java.io.InputStream;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Properties;
import java.util.Set;
import java.util.regex.Pattern;

import org.gudy.azureus2.plugins.utils.Formatters;
import org.gudy.azureus2.plugins.utils.StaticUtilities;

public class ExperimentParams {

	public static enum ISPMode {
		ISP_ENABLED,
		ISP_DISABLED,
		ISP_BY_ISPID,
	}

	public static enum ClientMode {
		CLIENT_ENABLED,
		CLIENT_DISABLED,
		CLIENT_BY_CLIENTID_BIT,
	}

	public static enum TorrentMode {
		TORRENT_ENABLED,
		TORRENT_DISABLED,
		TORRENT_BY_HASH_BIT,
		TORRENT_BY_HASH,
		TORRENT_BY_TRACKER,
	}

	private static final ISPMode DEFAULT_ISP_MODE = ISPMode.ISP_DISABLED;
	private static final ClientMode DEFAULT_CLIENT_MODE = ClientMode.CLIENT_DISABLED;
	private static final TorrentMode DEFAULT_TORRENT_MODE = TorrentMode.TORRENT_DISABLED;
	private static final String DEFAULT_TRACKER_HOST_REGEX = ".*";
	private static final Pattern DEFAULT_TRACKER_HOST_PATTERN = Pattern.compile(DEFAULT_TRACKER_HOST_REGEX);
	
	private static final Formatters FORMATTERS = StaticUtilities.getFormatters();

	private ISPMode ispMode_ = DEFAULT_ISP_MODE;
	public final ISPMode getISPMode() { return ispMode_; }

	private Set<String> enabledISPs_ = Collections.emptySet();
	public final boolean isISPEnabled(String isp) {
		return enabledISPs_.contains(isp);
	}

	private ClientMode clientMode_ = DEFAULT_CLIENT_MODE;
	public final ClientMode getClientMode() { return clientMode_; }

	private TorrentMode torrentMode_ = DEFAULT_TORRENT_MODE;
	public final TorrentMode getTorrentMode() { return torrentMode_; }

	private Set<String> torrentHashes_ = Collections.emptySet();
	public final boolean isTorrentHashEnabled(byte[] hash) {
		return torrentHashes_.contains(FORMATTERS.encodeBytesToString(hash).toLowerCase());
	}

	private String trackerHostRegex_ = DEFAULT_TRACKER_HOST_REGEX;
	private Pattern trackerHostPattern_ = DEFAULT_TRACKER_HOST_PATTERN;
	public final boolean isTrackerHostEnabled(String host) {
		return trackerHostPattern_.matcher(host).find();
	}

	public ExperimentParams() {
		/* Use defaults */
	}

	public ExperimentParams(Properties props) throws Exception {
		ispMode_ = ISPMode.valueOf(props.getProperty("ISPMode", DEFAULT_ISP_MODE.toString()));
		enabledISPs_ = new HashSet<String>(Arrays.asList(props.getProperty("EnabledISPs", "").split(",")));
		clientMode_ = ClientMode.valueOf(props.getProperty("ClientMode", DEFAULT_CLIENT_MODE.toString()));
		torrentMode_ = TorrentMode.valueOf(props.getProperty("TorrentMode", DEFAULT_TORRENT_MODE.toString()));
		torrentHashes_ = new HashSet<String>(Arrays.asList(props.getProperty("TorrentHashes", "").split(",")));
		trackerHostRegex_ = props.getProperty("TrackerHostPattern", DEFAULT_TRACKER_HOST_REGEX);
		trackerHostPattern_ = Pattern.compile(trackerHostRegex_);
	}

	public static ExperimentParams parse(InputStream stream) {
		try {
			Properties props = new Properties();

			/* Read the properties */
			props.load(stream);
			
			/* Try to construct using given properties */
			return new ExperimentParams(props);
		} catch (Exception e) {
			/* Fall back to default */
			return new ExperimentParams();
		}
	}
	
	private static final <T> boolean colEqual(Collection<T> c1, Collection<T> c2) {
		
		if (c1.size() != c2.size())
			return false;
		
		Iterator<T> i1 = c1.iterator();
		Iterator<T> i2 = c2.iterator();
		while (i1.hasNext())
			if (!i1.next().equals(i2.next()))
				return false;

		return true;
	}

	public boolean equals(Object o) {
		if (!(o instanceof ExperimentParams))
			return false;
		
		ExperimentParams oParams = (ExperimentParams)o;
		
		return ispMode_.equals(oParams.ispMode_)
			&& colEqual(enabledISPs_, oParams.enabledISPs_)
			&& clientMode_.equals(oParams.clientMode_)
			&& torrentMode_.equals(oParams.torrentMode_)
			&& colEqual(torrentHashes_, oParams.torrentHashes_)
			&& trackerHostRegex_.endsWith(oParams.trackerHostRegex_);
	}

}
