/**
 * P4P Vuze Plugin
 * 
 * Copyright 2008, Yale Laboratory of Networked Systems
 */

package net.yale.cs.p4p.p2p.azureus.client.listeners;

import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.Timer;
import java.util.TimerTask;

import net.yale.cs.p4p.p2p.azureus.client.P4PPlugin;
import net.yale.cs.p4p.p2p.azureus.client.data.PeerCache;
import net.yale.cs.p4p.p2p.azureus.client.data.QueryCache;
import net.yale.cs.p4p.p2p.azureus.client.data.TorrentState;
import net.yale.cs.p4p.p2p.azureus.client.data.Torrents;
import net.yale.cs.p4p.p2p.azureus.client.hacks.DiscoveredPeersHack;
import net.yale.cs.p4p.p2p.azureus.client.hacks.PeerClassifierHack;
import net.yale.cs.p4p.p2p.azureus.client.hacks.DiscoveredPeersHack.AnnouncePeer;

import org.gudy.azureus2.plugins.download.Download;
import org.gudy.azureus2.plugins.download.DownloadAnnounceResult;
import org.gudy.azureus2.plugins.download.DownloadAnnounceResultPeer;
import org.gudy.azureus2.plugins.download.DownloadScrapeResult;
import org.gudy.azureus2.plugins.download.DownloadTrackerListener;
import org.gudy.azureus2.plugins.peers.PeerDescriptor;
import org.gudy.azureus2.plugins.peers.PeerManager;
import org.gudy.azureus2.plugins.torrent.Torrent;

public class P4PDownloadTrackerListener implements DownloadTrackerListener
{

	private class PEXPeerQuery extends QueryCache {
		
		private class PEXPeerTask extends QueryTask {

			public PEXPeerTask(int interval) {
				super(interval);
			}

			private void processDownload(Download download) {
				/* Ignore if we are disabled */
				if (!plugin_.getConfig().getEnabled())
					return;
				
				Torrent torrent = download.getTorrent();
				if (torrent == null)
					return;
				
				PeerManager manager = download.getPeerManager();
				if (manager == null)
					return;
				
				/* Ignore if torrent is biased already or not P4P enabled */
				TorrentState torrentState = torrents_.getTorrent(torrent);
				if (torrentState.getVuzeBiased() || !plugin_.getExperiment().isP4PEnabled(download))
					return;

				List<PeerDescriptor> pexPeers = DiscoveredPeersHack.extractPEXPeers(manager); 
				if (pexPeers.isEmpty())
					return;
				
				plugin_.getLogger().log("Extracted " + pexPeers.size() + " PEX peers");
				handleNewPeers(download, torrents_.getTorrent(torrent), pexPeers);
			}

			protected void query() throws InterruptedException {
				for (Download download : plugin_.getPluginInterface().getDownloadManager().getDownloads())
					processDownload(download);
			}

		}

		public PEXPeerQuery(P4PPlugin plugin) {
			super(plugin, 1);
		}

		protected QueryTask makeQueryTask() {
			return new PEXPeerTask(plugin_.getConfig().getPEXCheckIntervalMS());
		}
		
	}

	private P4PPlugin plugin_;
	private PeerCache peerCache_;
	private Torrents torrents_;
	private PEXPeerQuery pexPeerQuery_;
	private Timer delayMonitorTimer_ = new Timer();

	public P4PDownloadTrackerListener(P4PPlugin plugin) {
		plugin_ = plugin;
		peerCache_ = plugin_.getPeerCache();
		torrents_ = plugin_.getTorrents();
		pexPeerQuery_ = new PEXPeerQuery(plugin_);
		pexPeerQuery_.start();
	}

	private void handleNewPeers(final Download download, TorrentState torrentState, Collection<PeerDescriptor> peers) {
		/* Add peers to the queue to have their PIDs looked up */
		boolean needsDelay = !peerCache_.addPeers(peers);

		/* Add to our global list of peers for this torrent */
		torrentState.peersDiscovered(peers, true);
		
		if (needsDelay) {
			/* Send notification to see if we need more peers.
			 * Notification is delayed so that we have some time to resolve PIDs for
			 * the peers we just discovered. */
			synchronized (delayMonitorTimer_) {
				delayMonitorTimer_.purge();
				delayMonitorTimer_.schedule(new TimerTask() {
								public void run() {
									plugin_.getLogger().log("Requesting delayed peer optimization");
									plugin_.getPeerMonitor().reschedule(download, true);
								}},
								plugin_.getConfig().getPeerScheduleDelay() * 1000);
			}
		} else {
			/* We were able to look up the peer locally; no need to wait */
			plugin_.getLogger().log("Requesting immediate peer optimization");
			plugin_.getPeerMonitor().reschedule(download, true);
		}
	}

	/*
	 * Intercept DHT and tracker peers.
	 */
	public void announceResult(DownloadAnnounceResult result) {
		
		/* Ignore if we are disabled */
		if (!plugin_.getConfig().getEnabled())
			return;
		
		if (result == null || result.getPeers() == null)
			return;

		Download download = result.getDownload();
		if (download == null)
			return;

		Torrent torrent = download.getTorrent();
		if (torrent == null)
			return;
		
		PeerManager peerManager = download.getPeerManager();
		if (peerManager == null)
			return;

		TorrentState torrentState = torrents_.getTorrent(torrent);

		/* Convert announce list into our own objects. Filter out peers that
		 * are reserved for Azureus since we don't want to remove them from
		 * the connection queue.
		 */
		Set<AnnouncePeer> announcePeers = new HashSet<AnnouncePeer>(result.getPeers().length);
		for (DownloadAnnounceResultPeer peer : result.getPeers()) {
			if (PeerClassifierHack.isAzureusIP(peer.getAddress())) {
				plugin_.getLogger().log(torrent, "Marking torrent as biased");
				torrentState.setVuzeBiased();
			}
			announcePeers.add(new AnnouncePeer(peer.getAddress(), peer.getPort(), peer.getUDPPort()));
		}

		/* Ignore announce if no peers */
		if (announcePeers.isEmpty()) {
			plugin_.getLogger().log(torrent, "Ignoring empty announce");
			return;
		}

		/* Ignore if torrent is biased already or not P4P enabled */
		if (torrentState.getVuzeBiased() || !plugin_.getExperiment().isP4PEnabled(download))
			return;

		plugin_.getLogger().log(torrent, "Received announce containing " + result.getPeers().length + " peers");
		
		/* We atomically do 3 things to the discovered peers queue in Vuze:
		 *   1) Remove peers that were just added during this announce
		 *   2) Refresh the set of queued peers that we've previously constructed (the ones
		 *      we scheduled before may have been kicked out by this tracker announce)
		 *   3) Return the set of peers that were removed from the queue (that is, the ones that
		 *      were contained in the announce) so we can add them to our master list of peers
		 * NOTE: It is very important that result.getPeers is a NON-EMPTY list.
		 *       Otherwise, the entire list of current peers in the queue will
		 *       be replaced. See setDiscoveredPeers() for more information. */
		List<PeerDescriptor> discoveredPeers = DiscoveredPeersHack.setDiscoveredPeers(peerManager, torrentState.getQueuedPeerDesc(), announcePeers);
		
		plugin_.getLogger().log(torrent, "Gathered " + discoveredPeers.size() + " peers from discovered queue");
		if (discoveredPeers.isEmpty())
			return;
		
		handleNewPeers(download, torrentState, discoveredPeers);
	}

	public void scrapeResult(DownloadScrapeResult result) { }

	public void cleanup() {
		pexPeerQuery_.interrupt();
		synchronized (delayMonitorTimer_) {
			delayMonitorTimer_.cancel();
		}
		
		pexPeerQuery_.joinNoThrow();
	}
	
}
