/*
 * Entry.java    1.0 2000/01/01
 *
 * Copyright (c) 2000 Stefan Nilsson
 * KTH, Nada
 * SE-100 44 Stockholm, Sweden
 * http://www.nada.kth.se/~snilsson
 *
 * The code presented in this file has been tested with care but
 * is not guaranteed for any purpose. The writer does not offer
 * any warranties nor does he accept any liabilities with respect
 * to the code.
 */

/*
 * Modifications made by Richard Alimi (richard.alimi@yale.edu)
 * for Vuze P4P Plugin developed by the Yale Laboratory of
 * Networked Systems.
 */

package se.kth.nada.router;

import net.yale.cs.p4p.api.data.PID;

/**
 * During the construction of the data structure the routing
 * table entries are stored in objects of this class.
 *
 * @author  Stefan Nilsson
 * @version 1.0, 1 jan 2000
 */
public class Entry implements Comparable {
    final int data; // the routing entry as a bit pattern
    final int length; // and its length
    final PID nexthop; // the corresponding next-hop

    public Entry(int data, int length, PID nexthop) {
        if (length < 0 || length > 32)
            throw new IllegalArgumentException("length out of bounds");

        // Clear the 32-len last data bits. This shouldn't be
        // necessary if the routing table is consistent.
        if (length == 0)
            this.data = 0; // >>>32 has no effect
        else
            this.data = data >>> (32-length) << (32-length);
        this.length = length;
        this.nexthop = nexthop;
    }

    /**
     * Tests if this entry is a prefix of t.
     *
     * @return <tt>true</tt> if this entry is a prefix of t;
     *         <tt>false</tt> otherwise.
     */
    public boolean isPrefixOf(Entry t) {
        return length == 0 || // >>>32 has no effect
               length <= t.length &&
               (data ^ t.data) >>> (32 - length) == 0;
    }

    /**
     * Compares two entries lexicographically.
     *
     * @param   o the Object to be compared.
     * @return  a negative integer, zero, or a positive integer as this object
     *		is less than, equal to, or greater than the specified object.
     * 
     * @throws ClassCastException if the specified object's type prevents it
     *         from being compared to this Object.
     */
    public int compareTo(Object o) {
        Entry e = (Entry) o;

        // Handle negativ numbers
        if (data < 0 && e.data >= 0)
            return 1;
        if (data >= 0 && e.data < 0)
            return -1;

        if (data < e.data)
            return -1;
        if (data > e.data)
            return 1;
        if (length < e.length)
            return -1;
        if (length > e.length)
            return 1;
        return 0;
    }

    /**
     * Compares this entry to the specified object.
     * The result is <tt>true</tt> if and only if the length and the
     * data fields are both equal.
     *
     * @param   anObject   the object to compare this <tt>Entry</tt>
     *                     against.
     * @return  <tt>true</tt> if the entries are equal;
     *          <tt>false</tt> otherwise.
     */
    public boolean equals(Object anObject) {
	  if (anObject != null && anObject instanceof Entry) {
	      Entry e = (Entry) anObject;
            if (e.data == data && e.length == length)
                return true;
	  }
	  return false;
    }

    /**
     * Returns a hashcode for this entry.  
     *
     * @return  a hash code value for this object. 
     */
    public int hashCode() {
        return data >>> (32 - length);
    }

    /**
     * Returns a string representing this entry.
     *
     * @return a string representing this entry.
     */
    public String toString() {
        StringBuffer sb = new StringBuffer(80);
        for (int i = 0; i < length; i++) {
            sb.append(data << i >>> 31);
            if (i%8 == 7 && i < length - 1)
                sb.append(" ");
        }
        sb.append(" (" + nexthop + ")");
        return sb.toString();
    }
}

